# Copyright 2010 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Iptables generator."""

import re
from string import Template  # pylint: disable=g-importing-member
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from absl import logging

from aerleon.lib import aclgenerator, nacaddr
from aerleon.lib.nacaddr import IPv4, IPv6
from aerleon.lib.policy import Policy, Term


class Term(aclgenerator.Term):
    """Generate Iptables policy terms."""

    # Validate that term does not contain any fields we do not
    # support.  This prevents us from thinking that our output is
    # correct in cases where we've omitted fields from term.
    _PLATFORM = 'iptables'
    _POSTJUMP_FORMAT = None
    _PREJUMP_FORMAT = Template('-A $filter -j $term')
    _TERM_FORMAT = Template('-N $term')
    _COMMENT_FORMAT = Template('-A $term -m comment --comment "$comment"')
    _FILTER_TOP_FORMAT = Template('-A $term')
    _LOG_FORMAT = Template('-j LOG --log-prefix $term')
    _PROTO_TABLE = {
        'icmpv6': '-p ipv6-icmp',
        'icmp': '-p icmp',
        'tcp': '-p tcp',
        'udp': '-p udp',
        'all': '-p all',
        'esp': '-p esp',
        'ah': '-p ah',
        'gre': '-p gre',
    }
    _TCP_FLAGS_TABLE = {
        'syn': 'SYN',
        'ack': 'ACK',
        'fin': 'FIN',
        'rst': 'RST',
        'urg': 'URG',
        'psh': 'PSH',
        'all': 'ALL',
        'none': 'NONE',
    }
    _KNOWN_OPTIONS_MATCHERS = {
        # '! -f' also matches non-fragmented packets.
        'first-fragment': '-m u32 --u32 4&0x3FFF=0x2000',
        'initial': '--syn',
        'tcp-initial': '--syn',
        'sample': '',
    }

    def __init__(
        self,
        term: Term,
        filter_name: str,
        trackstate: bool,
        filter_action: Optional[str],
        af: str = 'inet',
        verbose: bool = True,
        chained_terms: bool = True,
    ) -> None:
        """Setup a new term.

        Args:
          term: A policy.Term object to represent in iptables.
          filter_name: The name of the filter chan to attach the term to.
          trackstate: Specifies if conntrack should be used for new connections.
          filter_action: The default action of the filter.
          af: Which address family ('inet' or 'inet6') to apply the term to.
          verbose: boolean if comments should be printed

        Raises:
          UnsupportedFilterError: Filter is not supported.
        """
        super().__init__(term)
        self._action_table = {
            'accept': '-j ACCEPT',
            'deny': '-j DROP',
            'reject': '-j REJECT --reject-with icmp-host-prohibited',
            'reject-with-tcp-rst': '-j REJECT --reject-with tcp-reset',
            'next': '-j RETURN',
        }
        self.trackstate = trackstate
        self.term = term  # term object
        self.filter = filter_name  # actual name of filter
        self.default_action = filter_action
        self.options = []
        self.af = af
        self.verbose = verbose
        self.chained_terms = chained_terms
        if af == 'inet6':
            self._all_ips = nacaddr.IPv6('::/0')
            self._action_table['reject'] = '-j REJECT --reject-with ' 'icmp6-adm-prohibited'
        else:
            self._all_ips = nacaddr.IPv4('0.0.0.0/0')
            self._action_table['reject'] = '-j REJECT --reject-with ' 'icmp-host-prohibited'
        if self.chained_terms:
            self.term_name = '%s_%s' % (self.filter[:1], self.term.name)
        else:
            self.term_name = self.filter

    def __str__(self) -> str:
        ret_str = []

        # Don't render icmpv6 protocol terms under inet, or icmp under inet6
        if (self.af == 'inet6' and 'icmp' in self.term.protocol) or (
            self.af == 'inet' and 'icmpv6' in self.term.protocol
        ):
            logging.warning(
                self.NO_AF_LOG_PROTO.substitute(
                    term=self.term.name, proto=', '.join(self.term.protocol), af=self.af
                )
            )
            return ''

        # Term verbatim output - this will skip over most normal term
        # creation code by returning early. Warnings provided in policy.py
        if self.term.verbatim:
            for next_verbatim in self.term.verbatim:
                if next_verbatim[0] == self._PLATFORM:
                    ret_str.append(str(next_verbatim[1]))
            return '\n'.join(ret_str)

        # Create a new term
        self._SetDefaultAction()

        if self.chained_terms and self._TERM_FORMAT:
            ret_str.append(self._TERM_FORMAT.substitute(term=self.term_name))
        if self.chained_terms and self._PREJUMP_FORMAT:
            ret_str.append(
                self._PREJUMP_FORMAT.substitute(filter=self.filter, term=self.term_name)
            )

        if self.verbose:
            if self.term.owner:
                self.term.comment.append('Owner: %s' % self.term.owner)
            # reformat long comments, if needed
            #
            # iptables allows individual comments up to 256 chars.
            # But our generator will limit a single comment line to < 120, using:
            # max = 119 - 27 (static chars in comment command) - [length of term name]
            comment_max_width = 92 - len(self.term_name)
            if comment_max_width < 40:
                comment_max_width = 40
            comments = aclgenerator.WrapWords(self.term.comment, comment_max_width)
            # append comments to output
            if comments and comments[0]:
                for line in comments:
                    if not line:
                        continue  # iptables-restore does not like 0-length comments.
                    # term comments
                    # Strip out quotes as iptables cant have nested quotes
                    ret_str.append(
                        self._COMMENT_FORMAT.substitute(
                            filter=self.filter,
                            term=self.term_name,
                            comment=str(line).replace('\"', ''),
                        )
                    )

        # Unsupported configuration; in the case of 'accept' or 'next', we
        # skip the rule.  In other cases, we blow up (raise an exception)
        # to ensure that this is not considered valid configuration.
        if self.term.source_prefix or self.term.destination_prefix:
            if str(self.term.action[0]) not in set(['accept', 'next']):
                raise UnsupportedFilterError(
                    '%s %s %s %s %s %s %s %s'
                    % (
                        '\nTerm',
                        self.term.name,
                        'has action',
                        str(self.term.action[0]),
                        'with source_prefix or destination_prefix,',
                        ' which is unsupported in',
                        self._PLATFORM,
                        'iptables output.',
                    )
                )
            return '# skipped %s due to source or destination prefix rule' % self.term.name

        # protocol
        if self.term.protocol:
            protocol = self.term.protocol
        else:
            protocol = ['all']
        if 'hopopt' in protocol and self.af == 'inet':
            logging.warning('Term %s is using hopopt in IPv4 context.', self.term_name)
            return ''

        (term_saddr, exclude_saddr, term_daddr, exclude_daddr) = self._CalculateAddresses(
            self.term.source_address,
            self.term.source_address_exclude,
            self.term.destination_address,
            self.term.destination_address_exclude,
        )
        if not term_saddr:
            logging.warning(
                self.NO_AF_LOG_ADDR.substitute(term=self.term.name, direction='source', af=self.af)
            )
            return ''
        if not term_daddr:
            logging.warning(
                self.NO_AF_LOG_ADDR.substitute(
                    term=self.term.name, direction='destination', af=self.af
                )
            )
            return ''

        # ports
        source_port = []
        destination_port = []
        if self.term.source_port:
            source_port = self.term.source_port
        if self.term.destination_port:
            destination_port = self.term.destination_port

        # icmp-types
        icmp_types = ['']
        if self.term.icmp_type:
            icmp_types = self.NormalizeIcmpTypes(self.term.icmp_type, protocol, self.af)

        source_interface = ''
        if self.term.source_interface:
            source_interface = self.term.source_interface

        destination_interface = ''
        if self.term.destination_interface:
            destination_interface = self.term.destination_interface

        log_hits = False
        if self.term.logging:
            # Iptables sends logs to hosts configured syslog
            log_hits = True

        # options
        tcp_flags = []
        tcp_track_options = []
        for next_opt in [str(x) for x in self.term.option]:
            #
            # Sanity checking and high-ports are added as appropriate in
            # pre-processing that is done in __str__ within class Iptables.
            # Option established will add destination port high-ports if protocol
            # contains only tcp, udp or both.  This is done earlier in class Iptables.
            #
            if (
                next_opt.find('established') == 0 or next_opt.find('tcp-established') == 0
            ) and 'ESTABLISHED' not in [x.strip() for x in self.options]:
                if next_opt.find('tcp-established') == 0 and protocol != ['tcp']:
                    raise TcpEstablishedError(
                        '%s %s %s'
                        % (
                            '\noption tcp-established can only be applied for proto tcp.',
                            '\nError in term:',
                            self.term.name,
                        )
                    )

                if self.trackstate:
                    # Use nf_conntrack to track state -- works with any proto
                    self.options.append('-m state --state ESTABLISHED,RELATED')
                elif protocol == ['tcp']:
                    # Simple established-only rule for TCP: Must have ACK field
                    # (SYN/ACK or subsequent ACK), or RST and no other flags.
                    tcp_track_options = [
                        (['ACK'], ['ACK']),
                        (['SYN', 'FIN', 'ACK', 'RST'], ['RST']),
                    ]

            # Iterate through flags table, and create list of tcp-flags to append
            for next_flag in self._TCP_FLAGS_TABLE:
                if next_opt.find(next_flag) == 0:
                    tcp_flags.append(self._TCP_FLAGS_TABLE.get(next_flag))
            if next_opt in self._KNOWN_OPTIONS_MATCHERS:
                self.options.append(self._KNOWN_OPTIONS_MATCHERS[next_opt])
        if self.term.packet_length:
            # Policy format is "#-#", but iptables format is "#:#"
            self.options.append(
                '-m length --length %s' % self.term.packet_length.replace('-', ':')
            )
        if self.term.fragment_offset:
            self.options.append(
                '-m u32 --u32 4&0x1FFF=%s' % self.term.fragment_offset.replace('-', ':')
            )
        icmp_code = ['']
        if self.term.icmp_code:
            icmp_code = self.term.icmp_code

        for saddr in exclude_saddr:
            ret_str.extend(
                self._FormatPart(
                    '',
                    saddr,
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    self._action_table.get('next'),
                )
            )
        for daddr in exclude_daddr:
            ret_str.extend(
                self._FormatPart(
                    '',
                    '',
                    '',
                    daddr,
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    self._action_table.get('next'),
                )
            )

        for saddr in term_saddr:
            for daddr in term_daddr:
                for icmp in icmp_types:
                    for code in icmp_code:
                        for proto in protocol:
                            for tcp_matcher in tcp_track_options or (([], []),):
                                ret_str.extend(
                                    self._FormatPart(
                                        str(proto),
                                        saddr,
                                        source_port,
                                        daddr,
                                        destination_port,
                                        self.options,
                                        tcp_flags,
                                        icmp,
                                        code,
                                        tcp_matcher,
                                        source_interface,
                                        destination_interface,
                                        log_hits,
                                        self._action_table.get(str(self.term.action[0])),
                                    )
                                )

        if self._POSTJUMP_FORMAT and self.chained_terms:
            ret_str.append(
                self._POSTJUMP_FORMAT.substitute(filter=self.filter, term=self.term_name)
            )

        return '\n'.join(str(v) for v in ret_str if v)

    def _CalculateAddresses(
        self,
        term_saddr: List[Union[IPv4, IPv6]],
        exclude_saddr: List[Union[IPv4, IPv6]],
        term_daddr: List[Union[IPv4, IPv6]],
        exclude_daddr: List[Union[IPv4, IPv6]],
    ) -> Union[
        Tuple[
            List[Union[IPv4, IPv6]],
            List[Union[IPv4, IPv6]],
            List[Union[IPv4, IPv6]],
            List[Union[IPv4, IPv6]],
        ],
        Tuple[
            List[Union[IPv4, IPv6]],
            List[Union[IPv4, IPv6]],
            List[Union[IPv4, IPv6]],
            List[Union[IPv4, IPv6]],
        ],
        Tuple[
            List[Union[IPv4, IPv6]],
            List[Union[IPv4, IPv6]],
            List[Union[IPv4, IPv6]],
            List[Union[IPv4, IPv6]],
        ],
        Tuple[
            List[Union[IPv4, IPv6]],
            List[Union[IPv4, IPv6]],
            List[Union[IPv4, IPv6]],
            List[Union[IPv4, IPv6]],
        ],
    ]:
        """Calculate source and destination address list for a term.

        Args:
          term_saddr: source address list of the term
          exclude_saddr: source address exclude list of the term
          term_daddr: destination address list of the term
          exclude_daddr: destination address exclude list of the term

        Returns:
          tuple containing source address list, source exclude address list,
          destination address list, destination exclude address list in
          that order

        """
        # source address
        term_saddr_excluded = []
        if not term_saddr:
            term_saddr = [self._all_ips]
        if exclude_saddr:
            term_saddr_excluded.extend(nacaddr.ExcludeAddrs(term_saddr, exclude_saddr))

        # destination address
        term_daddr_excluded = []
        if not term_daddr:
            term_daddr = [self._all_ips]
        if exclude_daddr:
            term_daddr_excluded.extend(nacaddr.ExcludeAddrs(term_daddr, exclude_daddr))

        # Just to be safe, always have a result of at least 1 to avoid * by zero
        # returning incorrect results (10src*10dst=100, but 10src*0dst=0, not 10)
        bailout_count = (
            len(exclude_saddr)
            + len(exclude_daddr)
            + ((len(self.term.source_address) or 1) * (len(self.term.destination_address) or 1))
        )
        exclude_count = (len(term_saddr_excluded) or 1) * (len(term_daddr_excluded) or 1)

        # Use bailout jumps for excluded addresses if it results in fewer output
        # lines than nacaddr.ExcludeAddrs() method.
        if exclude_count < bailout_count:
            exclude_saddr = []
            exclude_daddr = []
            if term_saddr_excluded:
                term_saddr = term_saddr_excluded
            if term_daddr_excluded:
                term_daddr = term_daddr_excluded

        # With many sources and destinations, iptables needs to generate the
        # cartesian product of sources and destinations.  If there are no
        # exclude rules, this can instead be written as exclude [0/0 -
        # srcs], exclude [0/0 - dsts].
        v4_src_count = len([x for x in term_saddr if x.version == 4])
        v4_dst_count = len([x for x in term_daddr if x.version == 4])
        v6_src_count = len([x for x in term_saddr if x.version == 6])
        v6_dst_count = len([x for x in term_daddr if x.version == 6])
        num_pairs = v4_src_count * v4_dst_count + v6_src_count * v6_dst_count
        if num_pairs > 100:
            new_exclude_source = nacaddr.ExcludeAddrs([self._all_ips], term_saddr)
            new_exclude_dest = nacaddr.ExcludeAddrs([self._all_ips], term_daddr)
            # Invert the shortest list that does not already have exclude addresses
            if len(new_exclude_source) < len(new_exclude_dest) and not exclude_saddr:
                if len(new_exclude_source) + len(term_daddr) < num_pairs:
                    exclude_saddr = new_exclude_source
                    term_saddr = [self._all_ips]
            elif not exclude_daddr:
                if len(new_exclude_dest) + len(term_saddr) < num_pairs:
                    exclude_daddr = new_exclude_dest
                    term_daddr = [self._all_ips]
        term_saddr = [x for x in term_saddr if x.version == self.AF_MAP[self.af]]
        exclude_saddr = [x for x in exclude_saddr if x.version == self.AF_MAP[self.af]]
        term_daddr = [x for x in term_daddr if x.version == self.AF_MAP[self.af]]
        exclude_daddr = [x for x in exclude_daddr if x.version == self.AF_MAP[self.af]]
        return (term_saddr, exclude_saddr, term_daddr, exclude_daddr)

    def _FormatPart(
        self,
        protocol: str,
        saddr: Union[IPv6, IPv4, str],
        sport: Union[List[Tuple[int, int]], str],
        daddr: Union[IPv6, IPv4, str],
        dport: Union[List[Tuple[int, int]], str],
        options: Union[str, List[str]],
        tcp_flags: Union[str, List[str]],
        icmp_type: Union[int, str],
        code: Union[int, str],
        track_flags: Union[Tuple[List[str], List[str]], str],
        sint: str,
        dint: str,
        log_hits: Union[str, bool],
        action: str,
    ) -> List[str]:
        """Compose one iteration of the term parts into a string.

        Args:
          protocol: The network protocol
          saddr: Source IP address
          sport: Source port numbers
          daddr: Destination IP address
          dport: Destination port numbers
          options: Optional arguments to append to our rule
          tcp_flags: Which tcp_flag arguments, if any, should be appended
          icmp_type: What icmp protocol to allow, if any
          code: ICMP code allowed, if any
          track_flags: A tuple of ([check-flags], [set-flags]) arguments to tcp-flag
          sint: Optional source interface
          dint: Optional destination interface
          log_hits: Boolean, to log matches or not
          action: What should happen if this rule matches
        Returns:
          rval:  A single iptables argument line
        """
        src, dst = self._GenerateAddressStatement(saddr, daddr)

        filter_top = self._FILTER_TOP_FORMAT.substitute(filter=self.filter, term=self.term_name)

        source_int = ''
        if sint:
            source_int = '-i %s' % sint

        destination_int = ''
        if dint:
            destination_int = '-o %s' % dint

        log_jump = ''
        if log_hits:
            log_jump = self._LOG_FORMAT.substitute(term=self.term.name)
            if self.term.log_limit:
                log_jump = '-m --limit {}/{} {}'.format(
                    self.term.log_limit[0], self.term.log_limit[1], log_jump
                )

        if not options:
            options = []

        proto = self._PROTO_TABLE.get(str(protocol))
        # Don't drop protocol if we don't recognize it
        if protocol and not proto:
            proto = '-p %s' % str(protocol)

        if protocol == 'hopopt':
            proto = ''
            # Select 4 bytes at offset 0x3, mask out all but
            # last byte. That produces byte at position 7,
            # Next Header. Compare to 0x0, Hop By Hop
            options.append('-m u32 --u32 "0x3&0xff=0x0"')
        if protocol == 'fragment':
            proto = ''
            # Ditto, but compare to 0x2c, 44, Fragment
            options.append('-m u32 --u32 "0x3&0xff=0x2c"')

        # set conntrack state to NEW, unless policy requested "nostate"
        # for icmpv6 use trackstate only for type 128 and 139, as this is the
        # only supported types in nf_conntrack_proto_icmpv6
        if self.trackstate and (protocol != 'icmpv6' or icmp_type in [128, 139]):
            already_stateful = False
            # we will add new stateful arguments only if none already exist, such
            # as from "option:: established"
            for option in options:
                if 'state' in option:
                    already_stateful = True
            if not already_stateful:
                if 'ACCEPT' in action:
                    # We have to permit established/related since a policy may not
                    # have an existing blank permit for established/related, which
                    # may be more efficient, but slightly less secure.
                    options.append('-m state --state NEW,ESTABLISHED,RELATED')

        if tcp_flags or (track_flags and track_flags[0]):
            check_fields = ','.join(sorted(set(tcp_flags + track_flags[0])))
            set_fields = ','.join(sorted(set(tcp_flags + track_flags[1])))
            flags = '--tcp-flags %s %s' % (check_fields, set_fields)
        else:
            flags = ''

        icmp_type = str(icmp_type)
        if not icmp_type:
            icmp = ''
        elif str(protocol) == 'icmpv6':
            icmp = '-m icmp6 --icmpv6-type %s' % icmp_type
        else:
            icmp = '--icmp-type %s' % icmp_type
        if code:
            icmp += r'/%d' % code

        # format tcp and udp ports
        sports = dports = ['']
        if sport:
            sports = self._GeneratePortStatement(sport, source=True)
        if dport:
            dports = self._GeneratePortStatement(dport, dest=True)

        ret_lines = []
        for sport in sports:
            for dport in dports:
                rval = [filter_top]
                if re.search('multiport', sport) and not re.search('multiport', dport):
                    # Due to bug in iptables, use of multiport module before a single
                    # port specification will result in multiport trying to consume it.
                    # this is a little hack to ensure single ports are listed before
                    # any multiport specification.
                    dport, sport = sport, dport
                if str(protocol) == 'icmpv6':
                    # Due to a bug in ip6tables, iptables-save returns icmpv6 matches in
                    # order (address spec) (icmpv6 spec). Fake this using options
                    # datastructure.
                    options.extend((proto, icmp))
                    proto = ''
                    icmp = ''
                if str(self.af) == 'inet6' and str(protocol) == 'all' and 'REJECT' in str(action):
                    # Due to a bug in ip6tables, when -p all and -j REJECT, proto
                    # is being eaten
                    proto = ''
                for value in (
                    proto,
                    flags,
                    sport,
                    dport,
                    icmp,
                    src,
                    dst,
                    ' '.join(options),
                    source_int,
                    destination_int,
                ):
                    if value:
                        rval.append(str(value))
                if log_jump:
                    # -j LOG
                    ret_lines.append(' '.join(rval + [log_jump]))
                # -j ACTION
                ret_lines.append(' '.join(rval + [action]))
        return ret_lines

    def _GenerateAddressStatement(
        self, saddr: Union[IPv6, IPv4], daddr: Union[IPv6, IPv4]
    ) -> Tuple[str, str]:
        """Return the address section of an individual iptables rule.

        Args:
          saddr: source address of the rule
          daddr: destination address of the rule

        Returns:
          tuple containing source and destination address statement, in
          that order

        """
        src = ''
        dst = ''
        if not saddr or saddr == self._all_ips:
            src = ''
        else:
            src = '-s %s/%d' % (saddr.network_address, saddr.prefixlen)
        if not daddr or daddr == self._all_ips:
            dst = ''
        else:
            dst = '-d %s/%d' % (daddr.network_address, daddr.prefixlen)
        return (src, dst)

    def _GeneratePortStatement(
        self, ports: List[Tuple[int, int]], source: bool = False, dest: bool = False
    ) -> List[str]:
        """Return the 'port' section of an individual iptables rule.

        Args:
          ports: list of ports or port ranges (pairs)
          source: (bool) generate a source port rule
          dest: (bool) generate a dest port rule

        Returns:
          list holding the 'port' sections of an iptables rule.

        Raises:
          BadPortsError: if too many ports are passed in, or if both 'source'
                            and 'dest' are true.
          NotImplementedError: if both 'source' and 'dest' are true.
        """
        if not ports:
            return ''

        direction = ''  # default: no direction / '--port'.  As yet, unused.
        if source and dest:
            raise BadPortsError('_GeneratePortStatement called ambiguously.')
        elif source:
            direction = 's'  # source port / '--sport'
        elif dest:
            direction = 'd'  # dest port / '--dport'
        else:
            raise NotImplementedError('--port support not yet implemented.')

        # Normalize ports and get accurate port count.
        # iptables multiport module limits to 15, but we use 14 to ensure a range
        # doesn't tip us over the limit
        max_ports = 14
        norm_ports = []
        portstrings = []
        count = 0
        for port in ports:
            if port[0] == port[1]:
                norm_ports.append(str(port[0]))
                count += 1
            else:
                norm_ports.append('%d:%d' % (port[0], port[1]))
                count += 2
            if count >= max_ports:
                count = 0
                portstrings.append('-m multiport --%sports %s' % (direction, ','.join(norm_ports)))
                norm_ports = []
        if norm_ports:
            if len(norm_ports) == 1:
                portstrings.append('--%sport %s' % (direction, norm_ports[0]))
            else:
                portstrings.append('-m multiport --%sports %s' % (direction, ','.join(norm_ports)))
        return portstrings

    def _SetDefaultAction(self) -> None:
        """If term does not specify action, use filter default action."""
        if not self.term.action:
            self.term.action[0].value = self.default_action


class Iptables(aclgenerator.ACLGenerator):
    """Generates filters and terms from provided policy object."""

    _PLATFORM = 'iptables'
    _DEFAULT_PROTOCOL = 'all'
    SUFFIX = ''
    _RENDER_PREFIX = None
    _RENDER_SUFFIX = None
    _DEFAULTACTION_FORMAT = '-P %s %s'
    _DEFAULTACTION_FORMAT_CUSTOM_CHAIN = '-N %s'
    _DEFAULT_ACTION = 'DROP'
    _TERM = Term
    _TERM_MAX_LENGTH = 24
    _GOOD_FILTERS = ['INPUT', 'OUTPUT', 'FORWARD']
    _GOOD_OPTIONS = ['nostate', 'abbreviateterms', 'truncateterms', 'noverbose', 'nochainedterms']

    def __init__(self, pol: Policy, exp_info: int) -> None:
        self.iptables_policies = []
        super().__init__(pol, exp_info)

    def _BuildTokens(self) -> Tuple[Set[str], Dict[str, Set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {
            'counter',
            'destination_interface',
            'destination_prefix',
            'fragment_offset',
            'icmp_code',
            'logging',
            'log_limit',
            'owner',
            'packet_length',
            'routing_instance',
            'source_interface',
            'source_prefix',
        }

        supported_sub_tokens.update(
            {
                'option': {
                    'established',
                    'first-fragment',
                    'initial',
                    'sample',
                    'tcp-established',
                    'tcp-initial',
                    'syn',
                    'ack',
                    'fin',
                    'rst',
                    'urg',
                    'psh',
                    'all',
                    'none',
                },
            }
        )
        return supported_tokens, supported_sub_tokens

    def _WarnIfCustomTarget(self, target: str) -> None:
        """Emit a warning if a policy's default target is not a built-in chain."""
        if target not in self._GOOD_FILTERS:
            logging.warning(
                'Filter is generating a non-standard chain that will not '
                'apply to traffic unless linked from INPUT, OUTPUT or '
                'FORWARD filters. New chain name is: %s',
                target,
            )

    def _TranslatePolicy(self, pol: Policy, exp_info: int) -> None:
        """Translate a policy from objects into strings."""
        default_action = None
        good_default_actions = ['ACCEPT', 'DROP']
        good_afs = ['inet', 'inet6']
        chained_terms = True
        all_protocols_stateful = True
        self.verbose = True

        for header, terms in pol.filters:
            filter_type = None
            self.filter_options = header.FilterOptions(self._PLATFORM)[1:]
            filter_name = header.FilterName(self._PLATFORM)

            self._WarnIfCustomTarget(filter_name)

            # ensure all options after the filter name are expected
            for opt in self.filter_options:
                if opt not in good_default_actions + good_afs + self._GOOD_OPTIONS:
                    raise UnsupportedTargetOptionError(
                        '%s %s %s %s'
                        % (
                            '\nUnsupported option found in',
                            self._PLATFORM,
                            'target definition:',
                            opt,
                        )
                    )

            # disable stateful?
            if 'nostate' in self.filter_options:
                all_protocols_stateful = False
            if 'noverbose' in self.filter_options:
                self.verbose = False
            if 'nochainedterms' in self.filter_options:
                chained_terms = False

            # Check for matching af
            for address_family in good_afs:
                if address_family in self.filter_options:
                    # should not specify more than one AF in options
                    if filter_type is not None:
                        raise UnsupportedFilterError(
                            '%s %s %s %s'
                            % (
                                '\nMay only specify one of',
                                good_afs,
                                'in filter options:',
                                self.filter_options,
                            )
                        )
                    filter_type = address_family
            if filter_type is None:
                filter_type = 'inet'

            if self._PLATFORM == 'iptables' and filter_name == 'FORWARD':
                default_action = 'DROP'

            # does this policy override the default filter actions?
            for next_target in header.target:
                if next_target.platform == self._PLATFORM:
                    if len(next_target.options) > 1:
                        for arg in next_target.options:
                            if arg in good_default_actions:
                                default_action = arg
            if default_action and default_action not in good_default_actions:
                raise UnsupportedDefaultActionError(
                    '%s %s %s %s %s'
                    % (
                        '\nOnly',
                        ', '.join(good_default_actions),
                        'default filter action allowed;',
                        default_action,
                        'used.',
                    )
                )

            # add the terms
            new_terms = []
            term_names = set()
            for term in terms:
                term.name = self.FixTermLength(
                    term.name,
                    'abbreviateterms' in self.filter_options,
                    'truncateterms' in self.filter_options,
                )
                if term.name in term_names:
                    raise aclgenerator.DuplicateTermError(
                        'You have a duplicate term: %s' % term.name
                    )
                term_names.add(term.name)
                if not term.logging and term.log_limit:
                    raise LimitButNoLogError(
                        'Term %s: Cannoy use log-limit without logging' % term.name
                    )
                if not chained_terms:
                    term.name = filter_name

                term = self.FixHighPorts(
                    term, af=filter_type, all_protocols_stateful=all_protocols_stateful
                )
                if not term:
                    continue

                new_terms.append(
                    self._TERM(
                        term,
                        filter_name,
                        all_protocols_stateful,
                        default_action,
                        filter_type,
                        self.verbose,
                        chained_terms,
                    )
                )

            self.iptables_policies.append(
                (header, filter_name, filter_type, default_action, new_terms)
            )

    def SetTarget(self, target: str, action: Optional[str] = None) -> None:
        """Sets policy's target and default action.

        Args:
          target: (string) target name
          action: (string) default action, only valid if target is a built-in chain
        """
        # there is only one item in iptables_policies
        pol = list(self.iptables_policies[0])
        pol[1] = target
        self._WarnIfCustomTarget(target)
        if action:
            pol[3] = action
        self.iptables_policies[0] = tuple(pol)

    def __str__(self) -> str:
        target = []
        pretty_platform = '%s%s' % (self._PLATFORM[0].upper(), self._PLATFORM[1:])

        if self._RENDER_PREFIX:
            target.append(self._RENDER_PREFIX)

        for header, filter_name, filter_type, default_action, terms in self.iptables_policies:
            # Add comments for this filter
            target.append('# %s %s Policy' % (pretty_platform, header.FilterName(self._PLATFORM)))

            # reformat long text comments, if needed
            comments = aclgenerator.WrapWords(header.comment, 70)
            if comments and comments[0]:
                for line in comments:
                    target.append('# %s' % line)
                target.append('#')
            # add the p4 tags
            target.extend(aclgenerator.AddRepositoryTags('# '))
            target.append('# ' + filter_type)

            if filter_name in self._GOOD_FILTERS:
                if default_action:
                    target.append(self._DEFAULTACTION_FORMAT % (filter_name, default_action))
                elif self._PLATFORM == 'speedway':
                    # always specify the default filter states for speedway,
                    # if default action policy not specified for iptables, do nothing.
                    target.append(self._DEFAULTACTION_FORMAT % (filter_name, self._DEFAULT_ACTION))
            else:
                # Custom chains have no concept of default policy.
                target.append(self._DEFAULTACTION_FORMAT_CUSTOM_CHAIN % filter_name)
            # add the terms
            for term in terms:
                term_str = str(term)
                if term_str:
                    target.append(term_str)

        if self._RENDER_SUFFIX:
            target.append(self._RENDER_SUFFIX)

        target.append('')
        return '\n'.join(target)


class Error(aclgenerator.Error):
    """Base error class."""


class BadPortsError(Error):
    """Too many ports for a single iptables statement."""


class UnsupportedFilterError(Error):
    """Raised when we see an inappropriate filter."""


class NoIptablesPolicyError(Error):
    """Raised when a policy is received that doesn't support iptables."""


class TcpEstablishedError(Error):
    """Raised when a term has tcp-established option but not proto tcp only."""


class EstablishedError(Error):
    """Raised when a term has established option with inappropriate protocol."""


class UnsupportedDefaultActionError(Error):
    """Raised when a filter has an impermissible default action specified."""


class UnsupportedTargetOptionError(Error):
    """Raised when a filter has an impermissible default action specified."""


class LimitButNoLogError(Error):
    """Raised when log-limit is used by logging is not."""



# Copyright 2016 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Generic Windows security policy generator; requires subclassing."""

import string
from typing import Dict, List, Set, Tuple, Union

from absl import logging

from aerleon.lib import aclgenerator, nacaddr
from aerleon.lib.nacaddr import IPv4, IPv6
from aerleon.lib.policy import Header, Policy, Term

CMD_PREFIX = 'netsh ipsec static add '


class Term(aclgenerator.Term):
    """Generate generic windows policy terms."""

    _PLATFORM = 'windows'

    _COMMENT_FORMAT = string.Template(': $comment')

    # filter rules
    _ACTION_TABLE = {}

    def __init__(
        self, term: Term, filter_name: str, filter_action: None, af: str = 'inet'
    ) -> None:
        """Setup a new term.

        Args:
          term: A policy.Term object to represent in windows_ipsec.
          filter_name: The name of the filter chan to attach the term to.
          filter_action: The default action of the filter.
          af: Which address family ('inet' or 'inet6') to apply the term to.

        Raises:
          UnsupportedFilterError: Filter is not supported.
        """
        super().__init__(term)
        self.term = term  # term object
        self.filter = filter_name  # actual name of filter
        self.default_action = filter_action
        self.options = []
        self.af = af

        if af == 'inet6':
            self._all_ips = nacaddr.IPv6('::/0')
        else:
            self._all_ips = nacaddr.IPv4('0.0.0.0/0')

        self.term_name = '%s_%s' % (self.filter[:1], self.term.name)

    def __str__(self) -> str:
        ret_str = []

        # Don't render icmpv6 protocol terms under inet, or icmp under inet6
        if (self.af == 'inet6' and 'icmp' in self.term.protocol) or (
            self.af == 'inet' and 'icmpv6' in self.term.protocol
        ):
            logging.warning(
                self.NO_AF_LOG_PROTO.substitute(
                    term=self.term.name, proto=self.term.protocol, af=self.af
                )
            )
            return ''

        # append comments to output
        ret_str.append(
            self._COMMENT_FORMAT.substitute(
                filter=self.filter, term=self.term_name, comment=self.term.comment
            )
        )

        # if terms does not specify action, use filter default action
        if not self.term.action:
            self.term.action[0].value = self.default_action

        if self.term.action[0] == 'next':
            return ''

        if len(self.term.action) > 1:
            raise aclgenerator.UnsupportedFilterError(
                '\n%s %s %s %s'
                % (
                    'Multiple actions unsupported by',
                    self._PLATFORM,
                    '\nError in term:',
                    self.term.name,
                )
            )

        # protocol
        if self.term.protocol:
            protocols = self.term.protocol
        else:
            protocols = ['any']

        # addresses
        src_addr = self.term.source_address
        if not src_addr:
            src_addr = [self._all_ips]

        dst_addr = self.term.destination_address
        if not dst_addr:
            dst_addr = [self._all_ips]

        if self.term.source_address_exclude or self.term.destination_address_exclude:
            raise aclgenerator.UnsupportedFilterError(
                '\n%s %s %s %s'
                % (
                    'address exclusions unsupported by',
                    self._PLATFORM,
                    '\nError in term:',
                    self.term.name,
                )
            )

        # ports = Map the ports in a straight list since multiports aren't supported
        (src_ports, dst_ports) = self._HandlePorts(
            self.term.source_port, self.term.destination_port
        )

        # The windows ipsec driver requires either 'tcp' or 'udp' to be specified
        # if a srcport or dstport is specified.  Fail if src or dst ports are
        # specified and of the protocols are not exactly one or both of 'tcp'
        # or 'udp'.
        if (not set(protocols).issubset(set(['tcp', 'udp']))) and (
            len(src_ports) > 1 or len(dst_ports) > 1
        ):
            raise aclgenerator.UnsupportedFilterError(
                '%s %s %s'
                % (
                    '\n',
                    self.term.name,
                    'src or dst ports may only be specified with "tcp" and/or "udp".',
                )
            )

        # icmp-types
        (icmp_types, protocols) = self._HandleIcmpTypes(self.term.icmp_type, protocols)

        ret_str = []
        self._HandlePreRule(ret_str)
        self._CartesianProduct(
            src_addr, dst_addr, protocols, icmp_types, src_ports, dst_ports, ret_str
        )
        self._HandlePreRule(ret_str)

        return '\n'.join(str(v) for v in ret_str if v)

    def _HandleIcmpTypes(self, icmp_types: List[str], protocols: List[str]) -> Tuple[None, None]:
        """Perform implementation-specific icmp_type and protocol transforms.

        Note that icmp_types or protocols are passed as parameters in case they
        are to be munged prior to this function call, and may not be identical
        to self.term.* parameters.

        Args:
          icmp_types:  a list of icmp types, e.g., self.term.icmp_types
          protocols:  a list of protocols, e.g., self.term.protocols

        Returns:
          A pair of lists of (icmp_types, protocols)
        """
        return None, None

    def _HandlePorts(
        self, src_ports: List[Tuple[int, int]], dst_ports: List[Tuple[int, int]]
    ) -> Tuple[None, None]:
        """Perform implementation-specific port transforms.

        Note that icmp_types or protocols are passed as parameters in case they
        are to be munged prior to this function call, and may not be identical
        to self.term.* parameters.

        Args:
          src_ports:  list of source port range tuples, e.g., self.term.source_port
          dst_ports:  list of destination port range tuples

        Returns:
          A pair of lists of (icmp_types, protocols)
        """
        return None, None

    def _HandlePreRule(self, ret_str: List[str]) -> None:
        """Perform any pre-cartesian product transforms on the ret_str array.

        Args:
          ret_str:  an array of strings that will eventually be joined to form
            the string output for the term.
        """
        pass

    def _CartesianProduct(
        self,
        src_addr: List[Union[IPv4, IPv6]],
        dst_addr: List[Union[IPv4, IPv6]],
        protocol: List[str],
        unused_icmp_types: List[str],
        src_port: List[str],
        dst_port: List[str],
        ret_str: List[str],
    ) -> None:
        """Perform any the appropriate cartesian product of the input parameters.

        Args:
          src_addr: a type(IP) list of the source addresses
          dst_addr: a type(IP) list of the destination addresses
          protocol: a string list of the protocols
          icmp_types: a numeric list of the icmp_types
          src_ports: a (start, end) list of the source ports
          dst_ports: a (start,end) list of the destination ports
          ret_str:  an array of strings that will eventually be joined to form
            the string output for the term.
        """
        pass

    def _HandlePostRule(self, ret_str: str) -> None:
        """Perform any port-cartesian product transforms on the ret_str array.

        Args:
          ret_str:  an array of strings that will eventually be joined to form
            the string output for the term.
        """
        pass


class WindowsGenerator(aclgenerator.ACLGenerator):
    """Generates filters and terms from provided policy object."""

    _PLATFORM = 'windows'
    _DEFAULT_PROTOCOL = 'all'
    SUFFIX = '.bat'
    _RENDER_PREFIX = None
    _DEFAULT_ACTION = 'block'
    _TERM = Term

    _GOOD_AFS = ['inet', 'inet6']

    def _BuildTokens(self) -> Tuple[Set[str], Dict[str, Set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {'option'}
        supported_tokens -= {'verbatim'}

        supported_sub_tokens.update({'action': {'accept', 'deny'}})
        del supported_sub_tokens['option']
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: Policy, exp_info: int) -> None:
        """Translate a policy from objects into strings."""
        self.windows_policies = []
        default_action = None
        good_default_actions = ['permit', 'block']
        good_options = []

        for header, terms in pol.filters:
            filter_type = None
            filter_options = header.FilterOptions(self._PLATFORM)[1:]
            filter_name = header.FilterName(self._PLATFORM)

            # ensure all options after the filter name are expected
            for opt in filter_options:
                if opt not in good_default_actions + self._GOOD_AFS + good_options:
                    raise aclgenerator.UnsupportedTargetOptionError(
                        '%s %s %s %s'
                        % (
                            '\nUnsupported option found in',
                            self._PLATFORM,
                            'target definition:',
                            opt,
                        )
                    )

            # Check for matching af
            for address_family in self._GOOD_AFS:
                if address_family in filter_options:
                    # should not specify more than one AF in options
                    if filter_type is not None:
                        raise aclgenerator.UnsupportedFilterError(
                            '%s %s %s %s'
                            % (
                                '\nMay only specify one of',
                                self._GOOD_AFS,
                                'in filter options:',
                                filter_options,
                            )
                        )
                    filter_type = address_family
            if filter_type is None:
                filter_type = 'inet'

            # does this policy override the default filter actions?
            for next_target in header.target:
                if next_target.platform == self._PLATFORM:
                    if len(next_target.options) > 1:
                        for arg in next_target.options:
                            if arg in good_default_actions:
                                default_action = arg
            if default_action and default_action not in good_default_actions:
                raise aclgenerator.UnsupportedTargetOptionError(
                    '%s %s %s %s %s'
                    % (
                        '\nOnly',
                        ', '.join(good_default_actions),
                        'default filter action allowed;',
                        default_action,
                        'used.',
                    )
                )

            # add the terms
            new_terms = []
            term_names = set()
            for term in terms:
                if term.name in term_names:
                    raise aclgenerator.DuplicateTermError(
                        'You have a duplicate term: %s' % term.name
                    )
                term_names.add(term.name)

                if 'established' in term.option or 'tcp-established' in term.option:
                    continue
                new_terms.append(self._TERM(term, filter_name, default_action, filter_type))

            self.windows_policies.append(
                (header, filter_name, filter_type, default_action, new_terms)
            )

    def __str__(self) -> str:
        target = []
        pretty_platform = '%s%s' % (self._PLATFORM[0].upper(), self._PLATFORM[1:])

        if self._RENDER_PREFIX:
            target.append(self._RENDER_PREFIX)

        for header, _, filter_type, default_action, terms in self.windows_policies:
            # Add comments for this filter
            target.append(': %s %s Policy' % (pretty_platform, header.FilterName(self._PLATFORM)))

            self._HandlePolicyHeader(header, target)

            # reformat long text comments, if needed
            comments = aclgenerator.WrapWords(header.comment, 70)
            if comments and comments[0]:
                for line in comments:
                    target.append(': %s' % line)
                target.append(':')
            # add the p4 tags
            target.extend(aclgenerator.AddRepositoryTags(': '))
            target.append(': ' + filter_type)

            if default_action:
                raise aclgenerator.UnsupportedTargetOptionError(
                    'Windows generator does not support default actions'
                )

            # add the terms
            for term in terms:
                term_str = str(term)
                if term_str:
                    target.append(term_str)
                    self._HandleTermFooter(header, term, target)

        target.append('')
        return '\n'.join(target)

    def _HandlePolicyHeader(self, header: Header, target: List[str]) -> None:
        pass

    def _HandleTermFooter(self, header: Header, term: Term, target: List[str]) -> None:
        pass



# Copyright 2016 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Windows IP security policy generator."""


# pylint: disable=g-importing-member
from string import Template
from typing import Dict, List, Set, Tuple, Union

from absl import logging

from aerleon.lib import aclgenerator, windows
from aerleon.lib.nacaddr import IPv4, IPv6
from aerleon.lib.policy import Header


class Term(windows.Term):
    """Generate windows IP security policy terms."""

    _PLATFORM = 'windows_ipsec'
    CMD_PREFIX = 'netsh ipsec static add '

    # Windows IPSec Policy (which actually isn't limit to IPSec proper, as you
    # might expect) is structured such that you create:
    #   * One policy (more can be defined, but only one active)
    #   * One or more filter lists, which is similar to an IP chain (but does not
    #     support action:: next)
    #   * A filter, which is the matcher portion of a term
    #   * A filteraction, which is the action to perform when a a filter is
    #     matched.  Not that the filter action does not support logging, logging
    #     is associated with a auditpol filterlist.
    #   * a rule, which links a policy, filter list, and filter action.

    # Logging:
    # auditpol /set /subcategory:"Filtering Platform Packet Drop" /success:enable
    #   /failure:enable
    # auditpol /set /subcategory:"IPsec Driver" /success:enable /failure:enable
    # auditpol /set /subcategory:"IPsec Main Mode" /success:enable /failure:enable
    # auditpol /set /subcategory:"IPsec Quick Mode" /success:enable
    #   /failure:enable
    # auditpol /set /subcategory:"IPsec Extended Mode" /success:enable
    #   /failure:enable

    _MASK_ATOM = Template('${dir}mask=${mask}')
    _ADDR_ATOM = Template('${dir}addr=${addr} ${mask}')
    _PROTO_ATOM = Template('protocol=${protocol}')
    _PORT_ATOM = Template('${dir}port=${port}')

    # ipsec is not stateful, we need mirrored=yes to get responses:
    _FILTER_FORMAT = Template('filter filterlist=${name} mirrored=yes ' '${atoms}')
    _FILTERACTION_FORMAT = Template('filteraction name=${name} ' 'action=${action}')
    _FILTERLIST_FORMAT = Template('filterlist name=$name ')
    _RULE_FORMAT = Template(
        'rule name=$name policy=$policy ' 'filterlist=$filterlist ' 'filteraction=$filteraction '
    )
    _COMMENT_FORMAT = Template(': $comment')
    _LIST_SUFFIX = '-list'
    _ACTION_SUFFIX = '-action'

    # filter rules
    _ACTION_TABLE = {
        'accept': 'permit',
        'deny': 'block',
        'reject': 'block',
    }

    def _HandleIcmpTypes(
        self, icmp_types: List[str], protocols: List[str]
    ) -> Tuple[List[str], List[str]]:
        if icmp_types:
            raise aclgenerator.UnsupportedFilterError(
                '\n%s %s %s %s'
                % ('icmp types unsupported by', self._PLATFORM, '\nError in term:', self.term.name)
            )
        return ([''], protocols)

    def _HandlePorts(
        self, src_ports: List[Tuple[int, int]], dst_ports: List[Tuple[int, int]]
    ) -> Tuple[List[str], List[str]]:
        # ports = Map the ports in a straight list since multiports aren't supported
        return (self._CollapsePortTuples(src_ports), self._CollapsePortTuples(dst_ports))

    def _HandlePreRule(self, ret_str: List[str]) -> None:
        ret_str.append(self._ComposeFilterList())
        ret_str.append(self._ComposeFilterAction(self._ACTION_TABLE[self.term.action[0]]))

    def _CartesianProduct(
        self,
        src_addr: List[Union[IPv4, IPv6]],
        dst_addr: List[Union[IPv4, IPv6]],
        protocol: List[str],
        unused_icmp_types: List[str],
        src_port: List[str],
        dst_port: List[str],
        ret_str: List[str],
    ) -> None:
        for saddr in src_addr:
            if saddr.version != 4:
                logging.warning(
                    'WARNING: term contains a non IPv4 address %s, '
                    'ignoring element of term %s.',
                    saddr,
                    self.term_name,
                )
                continue

            for daddr in dst_addr:
                if daddr.version != 4:
                    logging.warning(
                        'WARNING: term contains a non IPv4 address %s, '
                        'ignoring element of term %s.',
                        daddr,
                        self.term_name,
                    )
                    continue

                for proto in protocol:
                    for sport in src_port:
                        for dport in dst_port:
                            ret_str.append(
                                self._ComposeFilter(
                                    saddr.network_address,
                                    daddr.network_address,
                                    proto,
                                    saddr.prefixlen,
                                    daddr.prefixlen,
                                    sport,
                                    dport,
                                )
                            )

    def _CollapsePortTuples(self, port_tuples: Tuple[int, int]) -> List[Union[str, int]]:
        ports = ['']
        for tpl in port_tuples:
            if tpl:
                (port_start, port_end) = tpl
                ports = list(range(port_start, port_end + 1))
        return ports

    def _ComposeFilterList(self) -> str:
        return self.CMD_PREFIX + self._FILTERLIST_FORMAT.substitute(
            name=self.term_name + self._LIST_SUFFIX
        )

    def _ComposeFilterAction(self, action) -> str:
        return self.CMD_PREFIX + self._FILTERACTION_FORMAT.substitute(
            name=self.term_name + self._ACTION_SUFFIX, action=action
        )

    def _ComposeFilter(
        self,
        srcaddr: Union[IPv4, IPv6],
        dstaddr: Union[IPv4, IPv6],
        proto: str,
        srcmask: int,
        dstmask: int,
        srcport: str,
        dstport: str,
    ):
        """Convert the given parameters to a netsh filter rule string."""
        atoms = []

        if srcmask == 0:
            atoms.append(self._ADDR_ATOM.substitute(dir='src', addr='any', mask=''))
        else:
            mask_atom = self._MASK_ATOM.substitute(dir='src', mask=srcmask)
            atoms.append(self._ADDR_ATOM.substitute(dir='src', addr=srcaddr, mask=mask_atom))

        if dstmask == 0:
            atoms.append(self._ADDR_ATOM.substitute(dir='dst', addr='any', mask=''))
        else:
            mask_atom = self._MASK_ATOM.substitute(dir='dst', mask=dstmask)
            atoms.append(self._ADDR_ATOM.substitute(dir='dst', addr=dstaddr, mask=mask_atom))

        if srcport:
            atoms.append(self._PORT_ATOM.substitute(dir='src', port=srcport))
        if dstport:
            atoms.append(self._PORT_ATOM.substitute(dir='dst', port=dstport))

        if proto:
            atoms.append(self._PROTO_ATOM.substitute(protocol=proto))

        return self.CMD_PREFIX + self._FILTER_FORMAT.substitute(
            name=self.term_name + self._LIST_SUFFIX, atoms=' '.join(atoms)
        )

    def ComposeRule(self, policy: str):
        return self.CMD_PREFIX + self._RULE_FORMAT.substitute(
            name=self.term_name + '-rule',
            policy=policy,
            filterlist=self.term_name + self._LIST_SUFFIX,
            filteraction=self.term_name + self._ACTION_SUFFIX,
        )


class WindowsIPSec(windows.WindowsGenerator):
    """Generates filters and terms from provided policy object."""

    _PLATFORM = 'windows_ipsec'
    _TERM = Term

    _POLICY_FORMAT = Template('policy name=$name assign=yes')
    _POLICY_SUFFIX = '-policy'

    _GOOD_AFS = ['inet']

    def _BuildTokens(self) -> Tuple[Set[str], Dict[str, Set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens -= {'icmp_type'}
        del supported_sub_tokens['icmp_type']
        return supported_tokens, supported_sub_tokens

    def _HandlePolicyHeader(self, header: Header, target: List[str]) -> None:
        policy_name = header.FilterName(self._PLATFORM) + self._POLICY_SUFFIX
        target.append(Term.CMD_PREFIX + self._POLICY_FORMAT.substitute(name=policy_name) + '\n')

    def _HandleTermFooter(self, header: Header, term: Term, target: List[str]):
        target.append(term.ComposeRule(header.FilterName(self._PLATFORM)) + '\n')



# Copyright 2015 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Google Compute Engine firewall generator.

More information about GCE networking and firewalls:
https://cloud.google.com/compute/docs/networking
https://cloud.google.com/compute/docs/reference/latest/firewalls
"""

import copy
import ipaddress
import json
import re
import sys
from typing import Dict, List, Set, Tuple, Union

from absl import logging

from aerleon.lib import gcp, nacaddr, policy

if sys.version_info < (3, 8):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


class Error(gcp.Error):
    """Generic error class."""


class GceFirewallError(Error):
    """Raised with problems in formatting for GCE firewall."""


class ExceededAttributeCountError(Error):
    """Raised when the total attribute count of a policy is above the maximum."""


LogConfig = TypedDict("LogConfig", {"enable": bool})
L4Matcher = TypedDict("L4Matcher", {"IPProtocol": str, "ports": "list[str]"})
FirewallRule = TypedDict(
    "FirewallRule",
    {
        "name": str,
        "description": str,
        "network": str,
        "priority": int,
        "sourceRanges": "list[str]",
        "destinationRanges": "list[str]",
        "sourceTags": "list[str]",
        "targetTags": "list[str]",
        "allowed": "list[L4Matcher]",
        "denied": "list[L4Matcher]",
        "direction": str,  # Actually Enum
        "logConfig": LogConfig,
    },
)


def IsDefaultDeny(term: policy.Term) -> bool:
    """Returns true if a term is a default deny without IPs, ports, etc."""
    skip_attrs = [
        'flattened',
        'flattened_addr',
        'flattened_saddr',
        'flattened_daddr',
        'action',
        'comment',
        'name',
        'logging',
    ]
    if 'deny' not in term.action:
        return False
    # This lc will look through all methods and attributes of the object.
    # It returns only the attributes that need to be looked at to determine if
    # this is a default deny.
    for i in [
        a
        for a in dir(term)
        if not a.startswith('__') and a.islower() and not callable(getattr(term, a))
    ]:
        if i in skip_attrs:
            continue
        v = getattr(term, i)
        if isinstance(v, str) and v:
            return False
        if isinstance(v, list) and v:
            return False

    return True


def GetNextPriority(priority: int) -> int:
    """Get the priority for the next rule."""
    return priority


class Term(gcp.Term):
    """Creates the term for the GCE firewall."""

    ACTION_MAP = {'accept': 'allowed', 'deny': 'denied'}
    # Restrict the number of addresses per term to 256.
    # Similar restrictions apply to source and target tags, and ports.
    # Details: https://cloud.google.com/vpc/docs/quota#per_network_2
    _TERM_ADDRESS_LIMIT = 256
    _TERM_SOURCE_TAGS_LIMIT = 30
    _TERM_TARGET_TAGS_LIMIT = 70
    _TERM_PORTS_LIMIT = 256

    # Firewall rule name has to match specific RE:
    # The first character must be a lowercase letter, and all following characters
    # must be a dash, lowercase letter, or digit, except the last character, which
    # cannot be a dash.
    # Details: https://cloud.google.com/compute/docs/reference/latest/firewalls
    _TERM_NAME_RE = re.compile(r'^[a-z]([-a-z0-9]*[a-z0-9])?$')

    # Protocols allowed by name from:
    # https://cloud.google.com/vpc/docs/firewalls#protocols_and_ports
    _ALLOW_PROTO_NAME = frozenset(
        [
            'tcp',
            'udp',
            'icmp',
            'esp',
            'ah',
            'ipip',
            'sctp',
            'all',  # Needed for default deny, do not use in policy file.
        ]
    )

    # Any protocol not in _ALLOW_PROTO_NAME must be passed by number.
    ALWAYS_PROTO_NUM = set(gcp.Term.PROTO_MAP.keys()) - _ALLOW_PROTO_NAME

    def __init__(
        self, term: policy.Term, inet_version: str = 'inet', policy_inet_version: str = 'inet'
    ) -> None:
        super().__init__(term)
        self.term = term
        self.inet_version = inet_version
        # This is to handle mixed, where the policy_inet_version is mixed,
        # but the term inet version is either inet/inet6.
        # This is only useful for term name and priority.
        self.policy_inet_version = policy_inet_version

        self._validateDirection()
        if self.term.source_address_exclude and not self.term.source_address:
            raise GceFirewallError(
                'GCE firewall does not support address exclusions without a source '
                'address list.'
            )
        # The reason for the error below isn't because of a GCE restriction, but
        # because we don't want to use a bad default of GCE that allows talking
        # to anything when there's no source address, source tag, or source service
        # account.
        if (
            not self.term.source_address and not self.term.source_tag
        ) and self.term.direction == 'INGRESS':
            raise GceFirewallError(
                'GCE firewall needs either to specify source address or source tags.'
            )
        if self.term.source_port:
            raise GceFirewallError('GCE firewall does not support source port restrictions.')
        if (
            self.term.source_address_exclude
            and self.term.source_address
            or self.term.destination_address_exclude
            and self.term.destination_address
        ):
            self.term.FlattenAll()
            if not self.term.source_address and self.term.direction == 'INGRESS':
                raise GceFirewallError(
                    'GCE firewall rule no longer contains any source addresses after '
                    'the prefixes in source_address_exclude were removed.'
                )
            # Similarly to the comment above, the reason for this error is also
            # because we do not want to use the bad default of GCE that allows for
            # talking to anything when there is no IP address provided for this field.
            if not self.term.destination_address and self.term.direction == 'EGRESS':
                raise GceFirewallError(
                    'GCE firewall rule no longer contains any destination addresses '
                    'after the prefixes in destination_address_exclude were removed.'
                )

    def __str__(self):
        """Convert term to a string."""
        json.dumps(self.ConvertToDict(), indent=2, separators=(',', ': '))

    def _validateDirection(self) -> None:
        if self.term.direction == 'INGRESS':
            if not self.term.source_address and not self.term.source_tag:
                raise GceFirewallError(
                    'Ingress rule missing required field oneof "sourceRanges" or ' '"sourceTags".'
                )

            if self.term.destination_address:
                raise GceFirewallError('Ingress rules cannot include ' '"destinationRanges.')

        elif self.term.direction == 'EGRESS':
            if self.term.source_address:
                raise GceFirewallError('Egress rules cannot include "sourceRanges".')

            if not self.term.destination_address:
                raise GceFirewallError('Egress rule missing required field "destinationRanges".')

            if self.term.destination_tag:
                raise GceFirewallError('GCE Egress rule cannot have destination tag.')

    def ConvertToDict(self) -> List[FirewallRule]:
        """Convert term to a dictionary.

        This is used to get a dictionary describing this term which can be
        output easily as a JSON blob.

        Returns:
          A dictionary that contains all fields necessary to create or update a GCE
          firewall.

        Raises:
          GceFirewallError: The term name is too long.
        """
        if self.term.owner:
            self.term.comment.append('Owner: %s' % self.term.owner)
        term_dict = {
            'description': ' '.join(self.term.comment),
            'name': self.term.name,
            'direction': self.term.direction,
        }
        if self.term.network:
            term_dict['network'] = self.term.network
            term_dict['name'] = '%s-%s' % (self.term.network.split('/')[-1], term_dict['name'])
        # Identify if this is inet6 processing for a term under a mixed policy.
        mixed_policy_inet6_term = False
        if self.policy_inet_version == 'mixed' and self.inet_version == 'inet6':
            mixed_policy_inet6_term = True
        # Update term name to have the IPv6 suffix for the inet6 rule.
        if mixed_policy_inet6_term:
            term_dict['name'] = gcp.GetIpv6TermName(term_dict['name'])

        # Checking counts of tags, and ports to see if they exceeded limits.
        if len(self.term.source_tag) > self._TERM_SOURCE_TAGS_LIMIT:
            raise GceFirewallError(
                'GCE firewall rule exceeded number of source tags per rule: %s' % self.term.name
            )
        if len(self.term.destination_tag) > self._TERM_TARGET_TAGS_LIMIT:
            raise GceFirewallError(
                'GCE firewall rule exceeded number of target tags per rule: %s' % self.term.name
            )

        if self.term.source_tag:
            if self.term.direction == 'INGRESS':
                term_dict['sourceTags'] = self.term.source_tag
            elif self.term.direction == 'EGRESS':
                term_dict['targetTags'] = self.term.source_tag
        if self.term.destination_tag and self.term.direction == 'INGRESS':
            term_dict['targetTags'] = self.term.destination_tag
        if self.term.priority:
            term_dict['priority'] = self.term.priority
            # Update term priority for the inet6 rule.
            if mixed_policy_inet6_term:
                term_dict['priority'] = GetNextPriority(term_dict['priority'])

        rules = []
        # If 'mixed' ends up in indvidual term inet_version, something has gone
        # horribly wrong. The only valid values are inet/inet6.
        term_af = self.AF_MAP.get(self.inet_version)
        if self.inet_version == 'mixed':
            raise GceFirewallError(
                'GCE firewall rule has incorrect inet_version for rule: %s' % self.term.name
            )

        # Exit early for inet6 processing of mixed rules that have only tags,
        # and no IP addresses, since this is handled in the inet processing.
        if mixed_policy_inet6_term:
            if not self.term.source_address and not self.term.destination_address:
                if 'targetTags' in term_dict or 'sourceTags' in term_dict:
                    return []

        saddrs = sorted(
            self.term.GetAddressOfVersion('source_address', term_af),
            key=ipaddress.get_mixed_type_key,
        )
        daddrs = sorted(
            self.term.GetAddressOfVersion('destination_address', term_af),
            key=ipaddress.get_mixed_type_key,
        )

        # If the address got filtered out and is empty due to address family, we
        # don't render the term. At this point of term processing, the direction
        # has already been validated, so we can just log and return empty rule.
        if self.term.source_address and not saddrs:
            logging.warning(
                'WARNING: Term %s is not being rendered for %s, '
                'because there are no addresses of that family.',
                self.term.name,
                self.inet_version,
            )
            return []
        if self.term.destination_address and not daddrs:
            logging.warning(
                'WARNING: Term %s is not being rendered for %s, '
                'because there are no addresses of that family.',
                self.term.name,
                self.inet_version,
            )
            return []

        filtered_protocols = []
        if not self.term.protocol:
            # Any protocol is represented as "all"
            filtered_protocols = ['all']
            logging.info(
                'INFO: Term %s has no protocol specified,'
                'which is interpreted as "all" protocols.',
                self.term.name,
            )

        proto_dict = copy.deepcopy(term_dict)

        if self.term.logging:
            proto_dict['logConfig'] = {'enable': True}

        for proto in self.term.protocol:
            # ICMP filtering by inet_version
            # Since each term has inet_version, 'mixed' is correctly processed here.
            # Convert protocol to number for uniformity of comparison.
            # PROTO_MAP always returns protocol number.
            if proto in self._ALLOW_PROTO_NAME:
                proto_num = self.PROTO_MAP[proto]
            else:
                proto_num = proto
            if proto_num == self.PROTO_MAP['icmp'] and self.inet_version == 'inet6':
                logging.warning(
                    'WARNING: Term %s is being rendered for inet6, ICMP '
                    'protocol will not be rendered.',
                    self.term.name,
                )
                continue
            if proto_num == self.PROTO_MAP['icmpv6'] and self.inet_version == 'inet':
                logging.warning(
                    'WARNING: Term %s is being rendered for inet, ICMPv6 '
                    'protocol will not be rendered.',
                    self.term.name,
                )
                continue
            if proto_num == self.PROTO_MAP['igmp'] and self.inet_version == 'inet6':
                logging.warning(
                    'WARNING: Term %s is being rendered for inet6, IGMP '
                    'protocol will not be rendered.',
                    self.term.name,
                )
                continue
            filtered_protocols.append(proto)
        # If there is no protocol left after ICMP/IGMP filtering, drop this term.
        if not filtered_protocols:
            return []
        for proto in filtered_protocols:
            # If the protocol name is not supported, protocol number is used.
            # This is done by default in policy.py.
            if proto not in self._ALLOW_PROTO_NAME:
                logging.info(
                    'INFO: Term %s is being rendered using protocol number', self.term.name
                )
            dest = {'IPProtocol': proto}

            if self.term.destination_port:
                ports = []
                for start, end in self.term.destination_port:
                    if start == end:
                        ports.append(str(start))
                    else:
                        ports.append('%d-%d' % (start, end))
                if len(ports) > self._TERM_PORTS_LIMIT:
                    raise GceFirewallError(
                        'GCE firewall rule exceeded number of ports per rule: %s' % self.term.name
                    )
                dest['ports'] = ports

            action = self.ACTION_MAP[self.term.action[0]]
            dict_val = []
            if action in proto_dict:
                dict_val = proto_dict[action]
                if not isinstance(dict_val, list):
                    dict_val = [dict_val]
            dict_val.append(dest)
            proto_dict[action] = dict_val

        # There's a limit of 256 addresses each term can contain.
        # If we're above that limit, we're breaking it down in more terms.
        if saddrs:
            source_addr_chunks = [
                saddrs[x : x + self._TERM_ADDRESS_LIMIT]
                for x in range(0, len(saddrs), self._TERM_ADDRESS_LIMIT)
            ]
            for i, chunk in enumerate(source_addr_chunks):
                rule = copy.deepcopy(proto_dict)
                if len(source_addr_chunks) > 1:
                    rule['name'] = '%s-%d' % (rule['name'], i + 1)
                rule['sourceRanges'] = [str(saddr) for saddr in chunk]
                rules.append(rule)
        elif daddrs:
            dest_addr_chunks = [
                daddrs[x : x + self._TERM_ADDRESS_LIMIT]
                for x in range(0, len(daddrs), self._TERM_ADDRESS_LIMIT)
            ]
            for i, chunk in enumerate(dest_addr_chunks):
                rule = copy.deepcopy(proto_dict)
                if len(dest_addr_chunks) > 1:
                    rule['name'] = '%s-%d' % (rule['name'], i + 1)
                rule['destinationRanges'] = [str(daddr) for daddr in chunk]
                rules.append(rule)
        else:
            rules.append(proto_dict)

        # Sanity checking term name lengths.
        long_rules = [rule['name'] for rule in rules if len(rule['name']) > 63]
        if long_rules:
            raise GceFirewallError('GCE firewall name ended up being too long: %s' % long_rules)
        return rules


class GCE(gcp.GCP):
    """A GCE firewall policy object."""

    _PLATFORM = 'gce'
    SUFFIX = '.gce'
    _SUPPORTED_AF = frozenset(('inet', 'inet6', 'mixed'))
    _ANY_IP = {
        'inet': nacaddr.IP('0.0.0.0/0'),
        'inet6': nacaddr.IP('::/0'),
    }
    # Supported is 63 but we need to account for dynamic updates when the term
    # is rendered (which can add proto and a counter).
    _TERM_MAX_LENGTH = 53
    _GOOD_DIRECTION = ['INGRESS', 'EGRESS']
    _OPTIONAL_SUPPORTED_KEYWORDS = set(['expiration', 'destination_tag', 'source_tag'])

    def _BuildTokens(self) -> Tuple[Set[str], Dict[str, Set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, _ = super()._BuildTokens()

        # add extra things
        supported_tokens |= {'destination_tag', 'expiration', 'owner', 'priority', 'source_tag'}

        # remove unsupported things
        supported_tokens -= {'icmp_type', 'verbatim'}
        # easier to make a new structure
        supported_sub_tokens = {'action': {'accept', 'deny'}}

        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int) -> None:
        self.gce_policies = []
        max_attribute_count = 0
        total_attribute_count = 0
        total_rule_count = 0

        for header, terms in pol.filters:
            filter_options = header.FilterOptions(self._PLATFORM)
            filter_name = header.FilterName(self._PLATFORM)

            network = ''
            direction = 'INGRESS'
            if filter_options:
                for i in self._GOOD_DIRECTION:
                    if i in filter_options:
                        direction = i
                        filter_options.remove(i)
            # Get the address family if set.
            address_family = 'inet'
            for i in self._SUPPORTED_AF:
                if i in filter_options:
                    address_family = i
                    filter_options.remove(i)

            for opt in filter_options:
                try:
                    max_attribute_count = int(opt)
                    logging.info('Checking policy for max attribute count %d', max_attribute_count)
                    filter_options.remove(opt)
                    break
                except ValueError:
                    continue

            if filter_options:
                network = filter_options[0]
            else:
                logging.warning('GCE filter does not specify a network.')

            if not terms:
                continue

            term_names = set()
            if IsDefaultDeny(terms[-1]):
                terms[-1].protocol = ['all']
                terms[-1].priority = 65534
                if direction == 'EGRESS':
                    if address_family != 'mixed':
                        # Default deny also gets processed as part of terms processing.
                        # The name and priority get updated there.
                        terms[-1].destination_address = [self._ANY_IP[address_family]]
                    else:
                        terms[-1].destination_address = [
                            self._ANY_IP['inet'],
                            self._ANY_IP['inet6'],
                        ]
                else:
                    if address_family != 'mixed':
                        terms[-1].source_address = [self._ANY_IP[address_family]]
                    else:
                        terms[-1].source_address = [self._ANY_IP['inet'], self._ANY_IP['inet6']]

            for term in terms:
                if term.stateless_reply:
                    logging.warning(
                        'WARNING: Term %s in policy %s is a stateless reply '
                        'term and will not be rendered.',
                        term.name,
                        filter_name,
                    )
                    continue
                term.network = network
                if not term.comment:
                    term.comment = header.comment
                if direction == 'EGRESS':
                    term.name += '-e'
                term.name = self.FixTermLength(term.name)
                if term.name in term_names:
                    raise GceFirewallError('Duplicate term name %s' % term.name)
                term_names.add(term.name)

                term.direction = direction
                if term.option:
                    raise GceFirewallError('GCE firewall does not support term options.')

                # Handle mixed for each indvidual term as inet and inet6.
                # inet/inet6 are treated the same.
                term_address_families = []
                if address_family == 'mixed':
                    term_address_families = ['inet', 'inet6']
                else:
                    term_address_families = [address_family]
                for term_af in term_address_families:
                    for rules in Term(term, term_af, address_family).ConvertToDict():
                        logging.debug(
                            'Attribute count of rule %s is: %d',
                            term.name,
                            GetAttributeCount(rules),
                        )
                        total_attribute_count += GetAttributeCount(rules)
                        total_rule_count += 1
                        if max_attribute_count and total_attribute_count > max_attribute_count:
                            # Stop processing rules as soon as the attribute count is over the
                            # limit.
                            raise ExceededAttributeCountError(
                                'Attribute count (%d) for %s exceeded the maximum (%d)'
                                % (total_attribute_count, filter_name, max_attribute_count)
                            )
                        self.gce_policies.append(rules)
        logging.info('Total rule count of policy %s is: %d', filter_name, total_rule_count)
        logging.info(
            'Total attribute count of policy %s is: %d', filter_name, total_attribute_count
        )

    def __str__(self) -> str:
        out = '%s\n\n' % (
            json.dumps(self.gce_policies, indent=2, separators=(',', ': '), sort_keys=True)
        )

        return out


def GetAttributeCount(dict_term: Dict[str, Union[List, str]]) -> int:
    """Calculate the attribute count of a term in its dictionary form.

    The attribute count of a rule is the sum of the number of ports, protocols, IP
    ranges, tags and target service account.

    Note: The goal of this function is not to determine if a term is valid, but
        to calculate its attribute count regardless of correctness.

    Args:
      dict_term: A dict object.

    Returns:
      int: The attribute count of the term.
    """
    addresses = len(dict_term.get('destinationRanges', [])) or len(
        dict_term.get('sourceRanges', [])
    )

    proto_ports = 0
    for allowed in dict_term.get('allowed', []):
        proto_ports += len(allowed.get('ports', [])) + 1  # 1 for ipProtocol
    for denied in dict_term.get('denied', []):
        proto_ports += len(denied.get('ports', [])) + 1  # 1 for ipProtocol

    tags = 0
    for _ in dict_term.get('sourceTags', []):
        tags += 1
    for _ in dict_term.get('targetTags', []):
        tags += 1

    service_accounts = 0
    for _ in dict_term.get('targetServiceAccount', []):
        service_accounts += 1

    return addresses + proto_ports + tags + service_accounts



# Copyright 2021 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Openconfig yang ACL generator.

More information about the Openconfig ACL model schema:
http://ops.openconfig.net/branches/models/master/openconfig-acl.html
"""

import copy
import json
import sys
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Set, Tuple, Union

from absl import logging

from aerleon.lib import aclgenerator, policy

if sys.version_info < (3, 8):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


class Error(aclgenerator.Error):
    """Generic error class."""


class OcFirewallError(Error):
    """Raised with problems in formatting for OpenConfig firewall."""


class ExceededAttributeCountError(Error):
    """Raised when the total attribute count of a policy is above the maximum."""


class TcpEstablishedWithNonTcpError(Error):
    """Raised when the TCP established option is set with a non TCP protocol."""


# Graceful handling of dict heierarchy for OpenConfig JSON.
def RecursiveDict() -> DefaultDict[Any, Any]:
    return defaultdict(RecursiveDict)


TransportConfig = TypedDict(
    "TransportConfig",
    {
        "source-port": Union[int, str],
        "destination-port": Union[int, str],
        "detail-mode": str,
        "builtin-detail": str,
    },
)
Transport = TypedDict("Transport", {"transport": TransportConfig})
IPConfig = TypedDict(
    "IPConfig", {"source-address": str, "destination-address": str, "protocol": int}
)
IP = TypedDict("IP", {"config": IPConfig})
ActionConfig = TypedDict("ActionConfig", {"forwarding-action": str})
Action = TypedDict("Action", {"config": ActionConfig})
ACLEntry = TypedDict(
    "ACLEntry",
    {"sequence-id": int, "actions": Action, "ipv4": IP, "ipv6": IP, "transport": Transport},
)
aclEntries = TypedDict("aclEntries", {"acl-entry": List[ACLEntry]})
ACLSetConfig = TypedDict("ACLSetConfig", {"name": str, "type": str})

ACLSet = TypedDict(
    "ACLSet", {"acl-entries": aclEntries, "config": ACLSetConfig, "name": str, "type": str}
)


class Term(aclgenerator.Term):
    """Creates the term for the OpenConfig firewall."""

    ACTION_MAP = {'accept': 'ACCEPT', 'deny': 'DROP', 'reject': 'REJECT'}

    # OpenConfig ip-protocols always will resolve to an 8-bit int, but these
    # common names are more convenient in a policy file.
    _ALLOW_PROTO_NAME = frozenset(['tcp', 'udp', 'icmp', 'esp', 'ah', 'ipip', 'sctp'])

    AF_RENAME = {
        4: 'ipv4',
        6: 'ipv6',
    }

    def __init__(self, term: policy.Term, inet_version: str = 'inet') -> None:
        super().__init__(term)
        self.term = term
        self.inet_version = inet_version

        # Combine (flatten) addresses with their exclusions into a resulting
        # flattened_saddr, flattened_daddr, flattened_addr.
        self.term.FlattenAll()

    def __str__(self) -> None:
        """Convert term to a string."""
        rules = self.ConvertToDict()
        json.dumps(rules, indent=2)

    def _tcp_established(self) -> Dict[str, str]:
        """Return's openconfig TCP_ESTABLISHED configuration.

        Other vendors (eg. SONiC) have slighly different implementations,
        This function permits inheritance."""
        return {'detail-mode': 'BUILTIN', 'builtin-detail': "TCP_ESTABLISHED"}

    def ConvertToDict(
        self,
    ) -> List[ACLEntry]:
        """Convert term to a dictionary.

        This is used to get a dictionary describing this term which can be
        output easily as an Openconfig JSON blob. It represents an "acl-entry"
        message from the OpenConfig ACL schema.

        Returns:
          A list of dictionaries that contains all fields necessary to create or
          update a OpenConfig acl-entry.
        """
        self.term_dict = RecursiveDict()

        # Rules will hold all exploded acl-entry dictionaries.
        rules = []

        # Convert the integer to the proper openconfig schema name str, ipv4/ipv6.
        term_af = self.AF_MAP.get(self.inet_version)
        family = self.AF_RENAME[term_af]

        self.SetName(self.term.name)

        # Action
        self.SetAction()

        # Ballot fatigue handling for 'any'.
        saddrs = self.term.GetAddressOfVersion('flattened_saddr', term_af)
        if not saddrs:
            saddrs = ['any']

        daddrs = self.term.GetAddressOfVersion('flattened_daddr', term_af)
        if not daddrs:
            daddrs = ['any']

        sports = self.term.source_port
        if not sports:
            sports = [(0, 0)]

        dports = self.term.destination_port
        if not dports:
            dports = [(0, 0)]

        protos = self.term.protocol
        if not protos:
            protos = ['none']

        self.term_dict = copy.deepcopy(self.term_dict)

        if self.term.comment:
            self.SetComments(self.term.comment)

        # Options
        self.SetOptions(family)

        # Source Addresses
        for saddr in saddrs:
            if saddr != 'any':
                self.SetSourceAddress(family, str(saddr))

            # Destination Addresses
            for daddr in daddrs:
                if daddr != 'any':
                    self.SetDestAddress(family, str(daddr))

                # Source Port
                for start, end in sports:
                    # 'any' starts and ends with zero.
                    if not start == end == 0:
                        self.SetSourcePorts(start, end)

                    # Destination Port
                    for start, end in dports:
                        if not start == end == 0:
                            self.SetDestPorts(start, end)

                        # Protocol
                        for proto in protos:
                            if isinstance(proto, str):
                                if proto != 'none':
                                    try:
                                        proto_num = self.PROTO_MAP[proto]
                                    except KeyError:
                                        raise OcFirewallError(
                                            'Protocol %s unknown. Use an integer.', proto
                                        )
                                    self.SetProtocol(family, proto_num)
                            else:
                                self.SetProtocol(family, proto)

                            # This is the business end of ace explosion.
                            # A dict is a reference type, so deepcopy is actually required.
                            rules.append(copy.deepcopy(self.term_dict))

        return rules

    def SetName(self, name: str) -> None:
        pass

    def SetAction(self) -> None:
        action = self.ACTION_MAP[self.term.action[0]]
        self.term_dict['actions'] = {}
        self.term_dict['actions']['config'] = {}
        self.term_dict['actions']['config']['forwarding-action'] = action

    def SetComments(self, comments: List[str]) -> None:
        pass

    def SetOptions(self, family: str) -> None:
        # options, 'family' unused
        if self.term.option:
            if 'tcp-established' in self.term.option:
                if self.term.protocol != ['tcp']:
                    raise TcpEstablishedWithNonTcpError(
                        f'tcp-established can only be used with tcp protocol in term {self.term.name}'
                    )
                self.term_dict['transport']['config'].update(self._tcp_established())

    def SetSourceAddress(self, family: str, saddr: str) -> None:
        self.term_dict[family]['config']['source-address'] = saddr

    def SetDestAddress(self, family: str, daddr: str) -> None:
        self.term_dict[family]['config']['destination-address'] = daddr

    def SetSourcePorts(self, start: int, end: int) -> None:
        if start == end:
            self.term_dict['transport']['config']['source-port'] = start
        else:
            self.term_dict['transport']['config']['source-port'] = '%d..%d' % (
                start,
                end,
            )

    def SetDestPorts(self, start: int, end: int) -> None:
        if start == end:
            self.term_dict['transport']['config']['destination-port'] = start
        else:
            self.term_dict['transport']['config']['destination-port'] = '%d..%d' % (
                start,
                end,
            )

    def SetProtocol(self, family: str, protocol: int) -> None:
        self.term_dict[family]['config']['protocol'] = protocol


class OpenConfig(aclgenerator.ACLGenerator):
    """A OpenConfig firewall policy object."""

    _PLATFORM = 'openconfig'
    SUFFIX = '.oacl'
    _SUPPORTED_AF = frozenset(('inet', 'inet6', 'mixed'))
    FAMILY_MAP = {'mixed': 'ACL_MIXED', 'inet6': 'ACL_IPV6', 'inet': 'ACL_IPV4'}
    _TERM = Term

    def _BuildTokens(self) -> Tuple[Set[str], Dict[str, Set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        # Remove unsupported things
        supported_tokens -= {'platform', 'platform_exclude', 'icmp-type', 'verbatim'}

        # OpenConfig ACL model only supports these three forwarding actions.
        supported_sub_tokens['action'] = {'accept', 'deny', 'reject'}

        supported_sub_tokens.update(
            {
                'option': {
                    'tcp-established',
                }
            }
        )

        return supported_tokens, supported_sub_tokens

    def _InitACLSet(self) -> None:
        """Initialize self.acl_sets with proper Typing"""
        self.acl_sets: List[ACLSet] = []

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int) -> None:
        self.total_rule_count = 0
        self._InitACLSet()

        for header, terms in pol.filters:
            filter_options = header.FilterOptions(self._PLATFORM)
            filter_name = header.FilterName(self._PLATFORM)

            # Options are anything after the platform name in the target message of
            # the policy header, [1:].

            # Get the address family if set.
            address_family = 'inet'
            for i in self._SUPPORTED_AF:
                if i in filter_options:
                    address_family = i
                    filter_options.remove(i)
            self._TranslateTerms(
                terms, address_family, filter_name, header.comment, filter_options
            )

        logging.info('Total rule count of policy %s is: %d', filter_name, self.total_rule_count)

    def _TranslateTerms(
        self,
        terms: List[Term],
        address_family: str,
        filter_name: str,
        hdr_comments: List[str],
        filter_options: List[str],
    ) -> None:
        """
        Factor out the translation of terms, such that it can be overridden by subclasses
        """
        oc_acl_entries: List[ACLEntry] = []

        for term in terms:
            # Handle mixed for each indvidual term as inet and inet6.
            # inet/inet6 are treated the same.
            term_address_families = []
            if address_family == 'mixed':
                term_address_families = ['inet', 'inet6']
            else:
                term_address_families = [address_family]
            for term_af in term_address_families:
                t = self._TERM(term, term_af)
                for rule in t.ConvertToDict():
                    self.total_rule_count += 1
                    rule['sequence-id'] = self.total_rule_count * 5
                    oc_acl_entries.append(rule)
        oc_type = self.FAMILY_MAP[address_family]
        oc_acl_set = {
            "acl-entries": {"acl-entry": oc_acl_entries},
            "config": {"name": filter_name, "type": oc_type},
            "name": filter_name,
            "type": oc_type,
        }
        self.acl_sets.append(oc_acl_set)

    def __str__(self) -> str:
        out = '%s\n\n' % (
            json.dumps(self.acl_sets, indent=2, separators=(',', ': '), sort_keys=True)
        )

        return out
